#include "Point.h"

#include "Transform.h"

#include <algorithm>
#include <cmath>
#include <numeric>

template<std::size_t DIM>
Point<DIM>::Point(const Point<2>& p)
    : Point(p.getX(), p.getY())
{

}

template<std::size_t DIM>
Point<DIM>::Point(double x, double y, double z)
{
    mCoords = std::vector<double>(DIM);
    if (DIM > 0)
    {
        mCoords[0] = x;
    }
    if (DIM > 1)
    {
        mCoords[1] = y;
    }
    if (DIM > 2)
    {
        mCoords[2] = z;
    }
}

template<std::size_t DIM>
Point<DIM>::Point(const Vector<double, DIM>& v)
{
    mCoords = std::vector<double>(DIM);
    for (std::size_t idx = 0; idx < DIM; idx++)
    {
        mCoords[idx] = v.getEntry(idx);
    }
}

template<std::size_t DIM>
Point<DIM>::Point(const DiscretePoint& point)
    : Point(static_cast<double>(point.getX()), static_cast<double>(point.getY()), 0.0)
{
    
}

template<std::size_t DIM>
Point<DIM>::Point(const Point<DIM>& reference, double offSetX, double offSetY, double offSetZ)
{
    mCoords = std::vector<double>(DIM);
    if (DIM > 0)
    {
        mCoords[0] = reference.getX() + offSetX;
    }
    if (DIM > 1)
    {
        mCoords[1] = reference.getY() + offSetY;
    }
    if (DIM > 2)
    {
        mCoords[2] = reference.getZ() + offSetZ;
    }
}

template<std::size_t DIM>
Point<DIM>::~Point()
{
};

template<std::size_t DIM>
double Point<DIM>::getX() const
{
    if (DIM > 0)
    {
        return mCoords[0];
    }
    else
    {
        return 0.0;
    }  
}

template<std::size_t DIM>
double Point<DIM>::getY() const
{
    if (DIM > 1)
    {
        return mCoords[1];
    }
    else
    {
        return 0.0;
    }
}

template<std::size_t DIM>
double Point<DIM>::getZ() const
{
    if (DIM > 2)
    {
        return mCoords[2];
    }
    else
    {
        return 0.0;
    }
}

template<std::size_t DIM>
double Point<DIM>::getDistance(const Point& point) const
{
    const auto deltaX = getDeltaX(point);
    const auto deltaY = getDeltaY(point);
    const auto deltaZ = getDeltaZ(point);
    return std::sqrt(deltaX* deltaX + deltaY* deltaY + deltaZ* deltaZ);
}

template<std::size_t DIM>
Vector<double, DIM> Point<DIM>::getDelta(const Point<DIM>& point) const
{
    return Vector<double, DIM>({ point.getX() - getX(), point.getY() - getY(), point.getZ() - getZ() });
}

template<std::size_t DIM>
double Point<DIM>::getDeltaX(const Point<DIM>& point) const
{
    return point.getX() - getX();
}

template<std::size_t DIM>
double Point<DIM>::getDeltaY(const Point<DIM>& point) const
{
    return point.getY() - getY();
}

template<std::size_t DIM>
double Point<DIM>::getDeltaZ(const Point<DIM>& point) const
{
    return point.getZ() - getZ();
}

template<std::size_t DIM>
bool Point<DIM>::isAtOrigin() const
{
    return std::accumulate(mCoords.begin(), mCoords.end(), 0.0) == 0.0;
}

template<std::size_t DIM>
Vector<double, DIM> Point<DIM>::getOriginOffset() const
{
    return Vector<double, DIM>(mCoords);
}

template<std::size_t DIM>
void Point<DIM>::apply(const Transform& transform)
{
    moveBy(-transform.getLocation().getX(), -transform.getLocation().getY(), -transform.getLocation().getZ());
    scale(transform.getScale().mX, transform.getScale().mY, transform.getScale().mZ);
}

template<std::size_t DIM>
void Point<DIM>::scale(double x, double y, double z)
{
    if (DIM > 0)
    {
        mCoords[0] *= x;
    }
    if (DIM > 1)
    {
        mCoords[1] *= y;
    }
    if (DIM > 2)
    {
        mCoords[2] *= z;
    }
}

template<std::size_t DIM>
void Point<DIM>::moveBy(double x, double y, double z)
{
    if (DIM > 0)
    {
        mCoords[0] += x;
    }
    if (DIM > 1)
    {
        mCoords[1] += y;
    }
    if (DIM > 2)
    {
        mCoords[2] += z;
    }
}

template<std::size_t DIM>
void Point<DIM>::moveBy(const Vector<double, DIM>& vector)
{
    for (std::size_t idx = 0; idx < DIM; idx++)
    {
        mCoords[idx] += vector.getEntry(idx);
    }
}

template<std::size_t DIM>
bool Point<DIM>::isEqual(const Point<DIM>& rhs) const
{
    return mCoords == rhs.mCoords;
}

template class Point<2>;
template class Point<3>;
