#include "PostscriptWriter.h"

#include "Arc.h"
#include "QuadraticBezierCurve.h"
#include "CubicBezierCurve.h"
#include "Line.h"
#include "LineSegment.h"

#include "PointParser.h"
#include "PathElement.h"
#include "Path.h"

void PostscriptWriter::setPrecision(std::size_t precision)
{
	mPrecision = precision;
}

std::string PostscriptWriter::toString(GeometryPath* path)
{
	std::string path_str;
	mCursor = { 0.0, 0.0 };
	for (const auto& feature : path->getFeatures())
	{
		auto start_offset = feature->getOffset();
		if (feature->getPathOffsetType() == PathOffsetType::RELATIVE_TO)
		{
			mCursor += start_offset;
			path_str += "m" + PointParser::toString(start_offset, " ", mPrecision);
		}
		else
		{
			mCursor = start_offset;
			path_str += "M" + PointParser::toString(start_offset, " ", mPrecision);
		}
		
		for (const auto& path_element : feature->getElements())
		{
			path_str += toString(path_element.get());
		}
		path_str += "Z ";
	}
	return path_str;
}

std::string PostscriptWriter::toString(PathElement<2>* pathElement)
{
	if (pathElement->getType() == AbstractGeometricItem::Type::CURVE)
	{
		if (auto curve = dynamic_cast<Curve<2>*>(pathElement); curve)
		{
			if (curve->getCurveType() == CurveType::ARC)
			{
				return toString(dynamic_cast<ntk::Arc*>(pathElement));
			}
			else if (curve->getCurveType() == CurveType::CUBIC_BEZIER)
			{
				return toString(dynamic_cast<CubicBezierCurve<2> *>(pathElement));
			}
			else if (curve->getCurveType() == CurveType::QUADRATIC_BEZIER)
			{
				return toString(dynamic_cast<QuadraticBezierCurve<2> *>(pathElement));
			}
		}
	}
	else if (pathElement->getType() == AbstractGeometricItem::Type::LINE)
	{
		if (auto line = dynamic_cast<Line<2>*>(pathElement); line)
		{
			return toString(line);
		}
	}
	else if (pathElement->getType() == AbstractGeometricItem::Type::LINE_SEGMENT)
	{
		if (auto line = dynamic_cast<LineSegment<2>*>(pathElement); line)
		{
			return toString(line);
		}
	}
	return {};
}

std::string PostscriptWriter::toString(QuadraticBezierCurve<2>* curve)
{
	Vector2 end_offset = curve->getEndOffset();
	std::string prefix = curve->isRelativeOffset() ? "q" : "Q";

	std::string result = prefix + PointParser::toString(curve->getControlOffset(), " ", mPrecision) + " " + PointParser::toString(end_offset, " ", mPrecision);
	if (curve->isRelativeOffset())
	{
		mCursor += end_offset;
	}
	else
	{
		mCursor = end_offset;
	}
	return result;
}

std::string PostscriptWriter::toString(ntk::Arc* arc)
{
	const auto large = arc->getUseLargeArc() ? "1" : "0";
	const auto sweep = arc->getSweepParam() ? "1" : "0";
	std::stringstream sstr;
	if (mPrecision > 0)
	{
		sstr.precision(mPrecision);
	}

	if (arc->isRelativeOffset())
	{
		sstr << "a";
	}
	else
	{
		sstr << "A";
	}
	sstr << arc->getRx() << " " << arc->getRy() << " " << arc->getRotation() << " " << large << " " << sweep << " ";

	auto end_offset = arc->getEndOffset();
	sstr << PointParser::toString(end_offset, " ", mPrecision);
	if (arc->isRelativeOffset())
	{	
		mCursor += end_offset;
	}
	else
	{
		mCursor = end_offset;
	}
	return sstr.str();
}

std::string PostscriptWriter::toString(CubicBezierCurve<2>* curve)
{
	Vector2 end_offset = curve->getEndOffset();
	std::string prefix = curve->isRelativeOffset() ? "c" : "C";

	const auto start_control = curve->getStartControlOffset();
	const auto end_control = curve->getStartControlOffset();
	const auto start_control_str = PointParser::toString(start_control, " ", mPrecision);
	const auto end_control_str = PointParser::toString(end_control, " ", mPrecision);
	const auto end_str = PointParser::toString(end_offset, " ", mPrecision);

	if (curve->isRelativeOffset())
	{
		mCursor += end_offset;
	}
	else
	{
		mCursor = end_offset;
	}
	return prefix + start_control_str + " " + end_control_str + " " + end_str;
}

std::string PostscriptWriter::toString(Line<2>* line)
{
	std::string prefix = line->isRelativeOffset() ? "l" : "L";
	std::string path = prefix;
	for (const auto& point : line->getPoints().getPoints())
	{
		path += PointParser::toString(point, " ", mPrecision) + " ";
	}

	if (line->isRelativeOffset())
	{
		mCursor += line->getEndOffset();
	}
	else
	{
		mCursor = line->getEndOffset();
	}
	return path;
}

std::string PostscriptWriter::toString(LineSegment<2>* line)
{
	if (line->isHorizontal())
	{
		double delta_x = line->isRelativeOffset() ? line->getPoint0().getDeltaX(line->getPoint1()) : line->getPoint1().getX();
		if (line->isRelativeOffset())
		{
			mCursor.moveBy(delta_x, 0.0);
			return "h" + PointParser::toString(delta_x, mPrecision);
		}
		else
		{
			mCursor = Vector2({ delta_x, 0.0 });
			return "H" + PointParser::toString(delta_x, mPrecision);
		}
	}
	else if (line->isVertical())
	{
		double delta_y = line->isRelativeOffset() ? line->getPoint0().getDeltaY(line->getPoint1()) : line->getPoint1().getY();
		if (line->isRelativeOffset())
		{
			mCursor.moveBy(delta_y, 0.0);
			return "v" + PointParser::toString(delta_y, mPrecision);
		}
		else
		{
			mCursor = Vector2({ 0.0, delta_y });
			return "V" + PointParser::toString(delta_y, mPrecision);
		}
	}
	else
	{
		if (line->isRelativeOffset())
		{
			mCursor += line->getEndOffset();
			return "l" + PointParser::toString(line->asVector(), " ", mPrecision);
		}
		else
		{
			mCursor = line->getPoint1();
			return "L" + PointParser::toString(line->getPoint1(), " ", mPrecision);
		}
	}
}